#include "../uwbgl_Defines.h"
#include "uwbgl_D3DRawMesh.h"

using namespace UWBGL_D3D;


RawMesh::RawMesh()
{

  m_Mesh = NULL;
  m_MeshMaterials = NULL;
  m_MeshTextures  = NULL;
  m_NumMaterials = 0;
  m_Device = NULL;
}

RawMesh::~RawMesh() 
{
  if (m_Device != NULL)
    ReleaseMemory();   
 
}

void RawMesh::ReleaseMemory()
{
    if( m_MeshMaterials != NULL ) {
        delete[] m_MeshMaterials;
        m_MeshMaterials = NULL;
    }

    if( m_MeshTextures != NULL) {

        for( unsigned int i = 0; i < m_NumMaterials; i++ ) {
          if( m_MeshTextures[i] != NULL )
            m_MeshTextures[i]->Release();
            m_MeshTextures[i] = NULL;
        }
        delete[] m_MeshTextures;
        m_MeshTextures = NULL;
        m_NumMaterials = 0;
    }

    if( m_Mesh != NULL ) {
        m_Mesh->Release();
        m_Mesh = NULL;
    }
}

//-----------------------------------------------------------------------------
// Load the mesh (in D3D .x format) and store into the buffer of the device
//
HRESULT RawMesh::LoadMesh(
  LPDIRECT3DDEVICE9 pDevice,
	ID3DXMatrixStack  *stack, 
  const char		*path,
  const char		*fileName
)
{
  LPD3DXBUFFER pD3DXMtrlBuffer;
  char     			fullPathName[200];

  if (m_Device != NULL)
    ReleaseMemory();

  m_Device = pDevice;
  m_Stack = stack;

    // Load the mesh from the specified file
  strcpy(fullPathName, path);
	strcat(fullPathName, "/");
	strcat(fullPathName, fileName);
  if( FAILED( D3DXLoadMeshFromX( fullPathName, D3DXMESH_SYSTEMMEM, 
                                   m_Device, NULL, 
                                   &pD3DXMtrlBuffer, NULL, &m_NumMaterials, 
                                   &m_Mesh ) ) )
  {
        return E_FAIL;
  }

  // We need to extract the material properties and texture names from the 
  // pD3DXMtrlBuffer
  D3DXMATERIAL* d3dxMaterials = (D3DXMATERIAL*)pD3DXMtrlBuffer->GetBufferPointer();
  m_MeshMaterials = new D3DMATERIAL9[m_NumMaterials];
  m_MeshTextures  = new LPDIRECT3DTEXTURE9[m_NumMaterials];

  for( unsigned int i=0; i<m_NumMaterials; i++ )
    {
        // Copy the material
        m_MeshMaterials[i] = d3dxMaterials[i].MatD3D;

        m_MeshTextures[i] = NULL;
        if (d3dxMaterials[i].pTextureFilename != NULL) {
            // Create the texture if there is one
			      strcpy(fullPathName, path);
	          strcat(fullPathName, "/");
	          strcat(fullPathName,d3dxMaterials[i].pTextureFilename);
            if( FAILED( D3DXCreateTextureFromFile( m_Device, 
                                                  fullPathName,
                                                  &m_MeshTextures[i] ) ) )
            {
                m_MeshTextures[i] = NULL;
            }
        }
         
    }

    // Done with the material buffer
    pD3DXMtrlBuffer->Release();
    
    
    //
	  // now compute the bounding sphere of this mesh 

  	LPDIRECT3DVERTEXBUFFER9 pVertexBuffer = NULL;
	  void*        pVertices;
    // Lock the vertex buffer, to generate a simple bounding sphere
    m_Mesh->GetVertexBuffer( &pVertexBuffer );
    pVertexBuffer->Lock( 0, 0, &pVertices, D3DLOCK_NOSYSLOCK );
   
    D3DXComputeBoundingSphere( (D3DXVECTOR3*)pVertices, m_Mesh->GetNumVertices(),
                                    D3DXGetFVFVertexSize(m_Mesh->GetFVF()),
                                    &m_ObjectCenter, &m_ObjectRadius );
    pVertexBuffer->Unlock();
	  pVertexBuffer->Release();

	  m_ObjectCenter = -m_ObjectCenter;
	  m_ObjectRadius = 1.0f/m_ObjectRadius;

    return S_OK;
}

//-----------------------------------------------------------------------------
VOID RawMesh::DrawOriginalSizeMesh(bool useColor, bool shouldTexture, const D3DXCOLOR &c, const D3DMATERIAL9 *useM)
{
    D3DMATERIAL9 m;
    memset(&m, 0, sizeof(D3DMATERIAL9));
    m.Ambient.a = c.a;
    m.Specular.a = c.a;
    m.Diffuse.a = c.a;
    m.Emissive = c;

    // now start drawing the regular mesh ...=
    m_Device->SetRenderState(D3DRS_LIGHTING, TRUE);
            // Mesh comes with material, so we should always be 
            // doing lighting

		// Meshes are divided into subsets, one for each material. Render them in
    // a loop
    for( unsigned int i=0; i<m_NumMaterials; i++ )
    {

        // Set the material and texture for this subset, either with the passed in color, or with the original mesh color
        if (useColor) {
            m_Device->SetMaterial(&m);
        } else {
						m_Device->SetMaterial( (useM==NULL) ? (&m_MeshMaterials[i]) : useM );
        }
        
        m_Device->SetTextureStageState(0, D3DTSS_COLOROP, D3DTOP_DISABLE);

        if ( shouldTexture && (m_MeshTextures[i] != NULL) ) {
            //
            // set up texture to be the color of the mesh
            //
            m_Device->SetTextureStageState( 0, D3DTSS_COLORARG1, D3DTA_TEXTURE );
            m_Device->SetTextureStageState( 0, D3DTSS_COLOROP,   D3DTOP_SELECTARG1);
            
            // Disable Texture coordinate transform
	          m_Device->SetTextureStageState(0 /* Texture Stage */, D3DTSS_TEXTURETRANSFORMFLAGS, 
								D3DTTFF_DISABLE /* Says Do not transform any more */);

            m_Device->SetSamplerState(0 /* Texture Stage */, D3DSAMP_MAGFILTER, D3DTEXF_NONE);
            m_Device->SetSamplerState(0 /* Texture Stage */, D3DSAMP_MINFILTER, D3DTEXF_NONE);
            m_Device->SetSamplerState(0 /* Texture Stage */, D3DSAMP_MIPFILTER, D3DTEXF_NONE);
            
            m_Device->SetSamplerState(0 /* Texture Stage */, D3DSAMP_ADDRESSU, D3DTADDRESS_WRAP);
            m_Device->SetSamplerState(0 /* Texture Stage */, D3DSAMP_ADDRESSV, D3DTADDRESS_WRAP);

            m_Device->SetTexture( 0, m_MeshTextures[i] );
        } 

        
        // Draw the mesh subset
        m_Mesh->DrawSubset( i );
    }
    //
    // Switch off texture
    //
    m_Device->SetTextureStageState(0, D3DTSS_COLOROP, D3DTOP_DISABLE);
    m_Device->SetRenderState(D3DRS_LIGHTING, FALSE);
}


//-----------------------------------------------------------------------------
VOID RawMesh::DrawNormalizedMesh(bool useColor, bool shouldTexture, const D3DXCOLOR &c, const D3DMATERIAL9 *useM)
{
    m_Stack->Push();
		m_Stack->ScaleLocal(m_ObjectRadius, m_ObjectRadius, m_ObjectRadius);
		m_Stack->TranslateLocal(m_ObjectCenter.x, m_ObjectCenter.y, m_ObjectCenter.z);

		m_Device->SetTransform(D3DTS_WORLD, m_Stack->GetTop());
    
            DrawOriginalSizeMesh(useColor, shouldTexture, c, useM);
    
	  m_Stack->Pop();
	  m_Device->SetTransform(D3DTS_WORLD, m_Stack->GetTop());
}
