#include "uwbgl_D3DXFormInfo.h"

//
// quaternion q is (x, y, z, w)
//
// to rotate a vertex v=(vx, vy, vz, 0)
//
// we need to:
//                      q v q*
//
// now if we have two successive rotations, then we have
//
//           q2 q1 v q1* q2*
//
//  let q3 = q2 q1
//
//  we see that 
//         q1* q2* = (q2 q1)* = q3*
//
//
// Meaning, we can remember successive rotations with quternions just like matrices
// notice, we must pre-concatenate successive's (instead of the post-concatenate like matrices)
//


using namespace UWBGL_D3D;

XFormInfo::XFormInfo()
{
  ResetTransformation();
  m_Pivot.x = m_Pivot.y = m_Pivot.z = 0.0f;
}

void XFormInfo::ResetTransformation()
{
  m_Translate.x = m_Translate.y = m_Translate.z = 0.0f;
  m_Scale.x = m_Scale.y = m_Scale.z = 1.0f;
  D3DXQuaternionIdentity(&m_Rotation);
  // 
  // To aovid changing of SceneNode behavior, Pivot is _NOT_ reset in the reset operation.
  // m_Pivot.x = m_Pivot.y = m_Pivot.z = 0.0f;
}

void XFormInfo::SetUpD3DXformStack(ID3DXMatrixStack *stack) const
{
  D3DXMATRIX m;
  GetOrientation(m);
        // This is a pretty expensive operation ...

  stack->TranslateLocal(m_Translate.x, m_Translate.y, m_Translate.z);

    stack->TranslateLocal(m_Pivot.x, m_Pivot.y, m_Pivot.z);
      stack->MultMatrixLocal(&m);
      stack->ScaleLocal(m_Scale.x, m_Scale.y, m_Scale.z);
    stack->TranslateLocal(-m_Pivot.x, -m_Pivot.y, -m_Pivot.z);
}


const D3DXVECTOR3& XFormInfo::XFormPivot() const
{
  return m_Pivot;
}

void XFormInfo::SetPivotTo(const D3DXVECTOR3 &p)
{
  m_Pivot = p;
}
void XFormInfo::SetPivotTo(float x, float y, float z)
{
  m_Pivot.x = x;
  m_Pivot.y = y;
  m_Pivot.z = z;
}
void XFormInfo::SetPivotXTo(float x) { m_Pivot.x = x; }
void XFormInfo::SetPivotYTo(float y) { m_Pivot.y = y; }
void XFormInfo::SetPivotZTo(float z) { m_Pivot.z = z; }

float XFormInfo::TranslateX() const { return m_Translate.x; }
float XFormInfo::TranslateY() const { return m_Translate.y; }
float XFormInfo::TranslateZ() const { return m_Translate.z; }
const D3DXVECTOR3& XFormInfo::Translation() const { return m_Translate; }

float XFormInfo::ScaleX() const { return m_Scale.x; }
float XFormInfo::ScaleY() const { return m_Scale.y; }
float XFormInfo::ScaleZ() const { return m_Scale.z; }
const D3DXVECTOR3& XFormInfo::Scale() const { return m_Scale; }

void XFormInfo::GetOrientation(D3DXMATRIX &m) const {  D3DXMatrixRotationQuaternion( &m, &m_Rotation); }
void XFormInfo::GetOrientationQ(D3DXQUATERNION &q) const { q = m_Rotation; }


void XFormInfo::SetTranslationTo(const D3DXVECTOR3 &d) {m_Translate = d; }
void XFormInfo::SetTranslationTo(float x, float y, float z)
{
  m_Translate.x = x;
  m_Translate.y = y;
  m_Translate.z = z;
}
void XFormInfo::SetTranslationXTo(float x) { m_Translate.x = x; }
void XFormInfo::SetTranslationYTo(float y) { m_Translate.y = y; }
void XFormInfo::SetTranslationZTo(float z) { m_Translate.z = z; }

void XFormInfo::UpdateTranslationBy(const D3DXVECTOR3 &v) { m_Translate += v; }
void XFormInfo::UpdateTranslationBy(float x, float y, float z) 
{
  m_Translate.x += x;
  m_Translate.y += y; 
  m_Translate.z += z; 
}
void XFormInfo::UpdateTranslationXBy(float x) { m_Translate.x += x; }
void XFormInfo::UpdateTranslationYBy(float y) { m_Translate.y += y; }
void XFormInfo::UpdateTranslationZBy(float z) { m_Translate.z += z; }

void XFormInfo::SetScaleTo(const D3DXVECTOR3 &s) { m_Scale = s; }
void XFormInfo::SetScaleTo(float s)
{
  m_Scale.x = s;
  m_Scale.y = s; 
  m_Scale.z = s;
}
void XFormInfo::SetScaleTo(float x, float y, float z) 
{
  m_Scale.x = x;
  m_Scale.y = y;
  m_Scale.z = z; 
}
void XFormInfo::SetScaleXTo(float x) { m_Scale.x = x; }
void XFormInfo::SetScaleYTo(float y) { m_Scale.y = y; }
void XFormInfo::SetScaleZTo(float z) { m_Scale.z = z; }

void XFormInfo::UpdateScaleBy(float s) 
{
  m_Scale.x *= s;
  m_Scale.y *= s;
  m_Scale.z *= s; 
}
void XFormInfo::UpdateScaleBy(float x, float y, float z) 
{
  m_Scale.x *= x;
  m_Scale.y *= y;
  m_Scale.z *= z; 
}
void XFormInfo::UpdateScaleXBy(float x) { m_Scale.x *= x; }
void XFormInfo::UpdateScaleYBy(float y) { m_Scale.y *= y; }
void XFormInfo::UpdateScaleZBy(float z) { m_Scale.z *= z; }


void XFormInfo::UpdateRotationBy(const D3DXQUATERNION &q)
{	
	// NOW, Quaternion shines, this is more efficient than concatenating two matrices
		D3DXQuaternionMultiply(&m_Rotation, &q, &m_Rotation);
}

void XFormInfo::UpdateRotationBy(const D3DXMATRIX &m)
{
  D3DXQUATERNION q;
  D3DXQuaternionRotationMatrix(&q, &m);
  UpdateRotationBy(q);
}

void XFormInfo::UpdateRotationXByRadian(float dx) 
{	
	D3DXQUATERNION q;
  D3DXVECTOR3 kXAxis(1.0f, 0.0f, 0.0f);
	D3DXQuaternionRotationAxis(&q, &kXAxis, dx);
	D3DXQuaternionMultiply(&m_Rotation, &q, &m_Rotation);
}

void XFormInfo::UpdateRotationYByRadian(float dy) 
{	
	D3DXQUATERNION q;
  D3DXVECTOR3 kYAxis(0.0f, 1.0f, 0.0f);
	D3DXQuaternionRotationAxis(&q, &kYAxis, dy);
	D3DXQuaternionMultiply(&m_Rotation, &q, &m_Rotation);
}

void XFormInfo::UpdateRotationZByRadian(float dz) 
{	
	D3DXQUATERNION q;
  D3DXVECTOR3 kZAxis(0.0f, 0.0f, 1.0f);
	D3DXQuaternionRotationAxis(&q, &kZAxis, dz);
	D3DXQuaternionMultiply(&m_Rotation, &q, &m_Rotation);
}

void XFormInfo::UpdateRotationXByDegree(float dx) 
{
	UpdateRotationXByRadian(D3DXToRadian(dx));
}

void XFormInfo::UpdateRotationYByDegree(float dy) 
{
	UpdateRotationYByRadian(D3DXToRadian(dy));
}

void XFormInfo::UpdateRotationZByDegree(float dz) 
{
	UpdateRotationZByRadian(D3DXToRadian(dz));
}

void XFormInfo::SetRotationToRadian(float x, float y, float z) 
{	
  D3DXQuaternionIdentity(&m_Rotation);
	UpdateRotationXByRadian(x);
	UpdateRotationYByRadian(y);
	UpdateRotationZByRadian(z);
}

void XFormInfo::SetRotationToDegree(float x, float y, float z) 
{	
  D3DXQuaternionIdentity(&m_Rotation);
	UpdateRotationXByRadian(D3DXToRadian(x));
	UpdateRotationYByRadian(D3DXToRadian(y));
	UpdateRotationZByRadian(D3DXToRadian(z));
}

void XFormInfo::SetRotationToRadian(D3DXVECTOR3 &r)
{	
	SetRotationToRadian(r.x, r.y, r.z);
}

void XFormInfo::SetRotationToDegree(D3DXVECTOR3 &r)
{	
	SetRotationToDegree(r.x, r.y, r.z);
}

void XFormInfo::SetOrientation(const D3DXMATRIX &m)
{
	D3DXQuaternionRotationMatrix(&m_Rotation, &m);
}

void XFormInfo::SetOrientationQ(const D3DXQUATERNION &q)
{
  m_Rotation = q;
}
